# -*- coding: utf-8 -*-
"""
    sphinxcontrib.websupport
    ~~~~~~~~~~~~~~~~~~~~~~~~

    A Python API to easily integrate Sphinx documentation into Web
    applications.

    :copyright: Copyright 2007-2018 by the Sphinx team, see README.
    :license: BSD, see LICENSE for details.
"""
from os import path

__import__('pkg_resources').declare_namespace(__name__)

package_dir = path.abspath(path.dirname(__file__))

from sphinxcontrib.websupport.core import WebSupport  # NOQA
from sphinxcontrib.websupport.version import __version__, __version_info__  # NOQA
