# -*- coding: utf-8 -*-
"""
    sphinxcontrib.websupport.version
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    :copyright: Copyright 2007-2018 by the Sphinx team, see README.
    :license: BSD, see LICENSE for details.
"""

__version__ = '1.2.3'
__version_info__ = tuple(map(int, __version__.split('.')))
